import json
from django.db import models
from django.forms import model_to_dict
import six


class DefaultContextSerializer(object):
    """
    Default class for serializing context data
    """

    def __init__(self, context):
        super(DefaultContextSerializer, self).__init__()
        self.context = context

    @property
    def data(self):
        """
        Data property that will return the serialized data
        :return:
        """

        # Create a serialized context dict
        serialized_context = self.serialize_value(self.context)

        # Return the serialized context
        return serialized_context

    def serialize_value(self, value):
        """
        Given a value, ensure that it is serialized properly
        :param value:
        :return:
        """
        # Create a list of serialize methods to run the value through
        serialize_methods = [
            self.serialize_model,
            self.serialize_json_string,
            self.serialize_list,
            self.serialize_dict
        ]

        # Run all of our serialize methods over our value
        for serialize_method in serialize_methods:
            value = serialize_method(value)

        # Return the serialized context value
        return value

    def serialize_model(self, value):
        """
        Serializes a model and all of its prefetched foreign keys
        :param value:
        :return:
        """

        # Check if the context value is a model
        if not isinstance(value, models.Model):
            return value

        # Serialize the model
        serialized_model = model_to_dict(value)

        # Check the model for cached foreign keys
        for model_field, model_value in serialized_model.items():
            cache_field = '_{0}_cache'.format(model_field)
            if hasattr(value, cache_field):
                serialized_model[model_field] = getattr(value, cache_field)

        # Return the serialized model
        return self.serialize_value(serialized_model)

    def serialize_json_string(self, value):
        """
        Tries to load an encoded json string back into an object
        :param json_string:
        :return:
        """

        # Check if the value might be a json string
        if not isinstance(value, six.string_types):
            return value

        # Make sure it starts with a brace
        if not value.startswith('{') or value.startswith('['):
            return value

        # Try to load the string
        try:
            return json.loads(value)
        except:
            return value

    def serialize_list(self, value):
        """
        Ensure that all values of a list or tuple are serialized
        :return:
        """

        # Check if this is a list or a tuple
        if not isinstance(value, (list, tuple)):
            return value

        # Loop over all the values and serialize the values
        return [
            self.serialize_value(list_value)
            for list_value in value
        ]

    def serialize_dict(self, value):
        """
        Ensure that all values of a dictionary are properly serialized
        :param value:
        :return:
        """

        # Check if this is a dict
        if not isinstance(value, dict):
            return value

        # Loop over all the values and serialize them
        return {
            dict_key: self.serialize_value(dict_value)
            for dict_key, dict_value in value.items()
        }



import json
from django.test.testcases import TransactionTestCase
from django_dynamic_fixture import G
from mock import patch, call

from entity_event.context_serializer import DefaultContextSerializer


class DefaultContextSerializerTests(TransactionTestCase):
    def setUp(self):
        super(DefaultContextSerializerTests, self).setUp()

        # Create some fake context to work with
        self.context = dict(
            test='test'
        )

        # Create a serializer to test with
        self.serializer = DefaultContextSerializer(self.context)

    @patch.object(DefaultContextSerializer, 'serialize_value', autospec=True)
    def test_data_property(self, mock_serialize_value):
        # Call the property
        response = self.serializer.data

        # Assert that we have a proper response
        self.assertEqual(response, mock_serialize_value.return_value)

    @patch.object(DefaultContextSerializer, 'serialize_model', autospec=True)
    @patch.object(DefaultContextSerializer, 'serialize_json_string', autospec=True)
    @patch.object(DefaultContextSerializer, 'serialize_list', autospec=True)
    @patch.object(DefaultContextSerializer, 'serialize_dict', autospec=True)
    def test_serialize_value(self, *serialize_methods):
        # Setup the return values of each method
        for serialize_method in serialize_methods:
            serialize_method.return_value = self.context

        # Call the method
        response = self.serializer.serialize_value(self.context)

        # Assert we have a proper response
        self.assertEqual(response, serialize_methods[0].return_value)

        # Assert that each serialize method was called properly
        for serialize_method in serialize_methods:
            serialize_method.assert_called_once_with(self.serializer, self.context)

    def test_serialize_model_non_model(self):
        # Call the method
        response = self.serializer.serialize_model('test')

        # Assert we have a proper response
        self.assertEqual(response, 'test')

    def test_serialize_model(self):
        from entity_event.tests.models import TestModel

        # Create a model to test with
        model = G(TestModel)

        # Fetch the model so we dont have the fks loaded and only select one related
        model = TestModel.objects.select_related('fk').get(id=model.id)

        # Call the method
        response = self.serializer.serialize_model(model)

        # Assert that we have a proper response
        self.assertEqual(
            response,
            {
                'fk_m2m': [],
                'fk2': model.fk2.id,
                'fk': {
                    'id': model.fk.id,
                    'value': model.fk.value
                },
                'id': model.id,
                'value': model.value
            }
        )

    def test_serialize_json_string_non_string(self):
        # Call the method
        response = self.serializer.serialize_json_string(dict())

        # Assert we have a proper response
        self.assertEqual(response, dict())

    def test_serialize_json_string_non_json_string(self):
        # Call the method
        response = self.serializer.serialize_json_string('test')

        # Assert we have a proper response
        self.assertEqual(response, 'test')

    def test_serialize_json_string_bad_json_string(self):
        # Call the method
        response = self.serializer.serialize_json_string('{test')

        # Assert we have a proper response
        self.assertEqual(response, '{test')

    def test_serialize_json_string(self):
        # Create a json string to test
        test_dict = dict(test='test')
        test_json = json.dumps(test_dict)

        # Call the method
        response = self.serializer.serialize_json_string(test_json)

        # Assert that we have a proper response
        self.assertEqual(
            response,
            test_dict
        )

    def test_serialize_list_non_list(self):
        # Call the method
        response = self.serializer.serialize_list('test')

        # Assert we have a proper response
        self.assertEqual(response, 'test')

    @patch.object(DefaultContextSerializer, 'serialize_value', autospec=True)
    def test_serialize_list_list(self, mock_serialize_value):
        # Setup a test list
        test_list = ['one', 'two', 'three']

        # Call the method
        response = self.serializer.serialize_list(test_list)

        # Assert that we have the proper response
        self.assertEqual(
            response,
            [
                mock_serialize_value.return_value,
                mock_serialize_value.return_value,
                mock_serialize_value.return_value,
            ]
        )

        # Assert that we called serialize value on on values of the list
        self.assertEqual(
            mock_serialize_value.mock_calls,
            [
                call(self.serializer, 'one'),
                call(self.serializer, 'two'),
                call(self.serializer, 'three'),
            ]
        )

    @patch.object(DefaultContextSerializer, 'serialize_value', autospec=True)
    def test_serialize_list_tuple(self, mock_serialize_value):
        # Setup a test tuple
        test_tuple = ('one', 'two', 'three')

        # Call the method
        response = self.serializer.serialize_list(test_tuple)

        # Assert that we have the proper response
        self.assertEqual(
            response,
            [
                mock_serialize_value.return_value,
                mock_serialize_value.return_value,
                mock_serialize_value.return_value,
            ]
        )

        # Assert that we called serialize value on on values of the list
        self.assertEqual(
            mock_serialize_value.mock_calls,
            [
                call(self.serializer, 'one'),
                call(self.serializer, 'two'),
                call(self.serializer, 'three'),
            ]
        )

    def test_serialize_dict_non_dict(self):
        # Call the method
        response = self.serializer.serialize_dict('test')

        # Assert we have a proper response
        self.assertEqual(response, 'test')

    @patch.object(DefaultContextSerializer, 'serialize_value', autospec=True)
    def test_serialize_dict(self, mock_serialize_value):
        # Setup a test dict
        test_dict = dict(one='one', two='two')

        # Call the method
        response = self.serializer.serialize_dict(test_dict)

        # Assert we have a proper response
        self.assertEqual(
            response,
            dict(
                one=mock_serialize_value.return_value,
                two=mock_serialize_value.return_value,
            )
        )

        # Assert that we called serialize value on on values of the dict
        mock_serialize_value.assert_has_calls([
            call(self.serializer, 'one'),
            call(self.serializer, 'two'),
        ], any_order=True)
